from __future__ import annotations
from typing import Set, FrozenSet, List, Dict, Union, Tuple, Deque, Optional, TypeVar, Any
from enum import Enum
import enum
from copy import deepcopy
from collections import deque


class Composition(Enum):
    Union = enum.auto()
    Intersection = enum.auto()
    Subtraction = enum.auto()


class IsEmptyResult:
  def __init__(self, counterexample : Optional[...] = None):
    self.counterexample = counterexample

  def __bool__(self):
    return self.counterexample is None

class IsEquivalentResult:
  def __init__(self, left_counterexample : Optional[...] = None, right_counterexample : Optional[...] = None):
    self.left_counterexample = left_counterexample
    self.right_counterexample = right_counterexample

  def __bool__(self):
    return self.left_counterexample is None and self.right_counterexample is None

class Character:
    def __init__(self, name: str):
        self.name = name

    def __eq__(self, obj):
        if isinstance(obj, Character):
            return obj.name == self.name
        return False

    def __hash__(self):
        return hash(self.name)


class Eps:
    def __init__(self):
        pass

    def __eq__(self, other):
        return isinstance(other, Eps)

    def __hash__(self):
        return hash("epsilon")

class State:
    def __init__(self, name: str):
        self.name = name

    def __eq__(self, obj):
        if isinstance(obj, State):
            return obj.name == self.name
        return False

    def __hash__(self):
        return hash(self.name)


class DFA:
    Transition = Dict[Tuple[State, Character], State]
    type_var = TypeVar('type_var')

    def __init__(self, states: Set[State],
                   characters: Set[Character],
                   transition: Transition,
                         init: State,
                        final: Set[State]):

        self.states = states
        self.characters = characters
        self.transition = transition
        self.init = init
        self.final = final
        assert self.check()

    def check(self) -> bool:
        assert len(self.states) > 0, "empty automaton"

        for (state, character) in self.transition:
            assert state in self.states, "unknown state " + state.name
            assert character in self.characters, "unknown character " + character.name
            assert self.transition[state, character] in self.states, "unknown state " + self.transition[
                state, character].name

        assert self.init in self.states, "init not in states"
        for state in self.final:
            assert state in self.states, "unknown state " + state.name

        return True

    def total(self) -> DFA:
        if self.is_total():
            return self

        dfa = deepcopy(self)
        total: Dict[Tuple[State, Character], State] = {}

        # check: consider import itertools
        hell_id = 0
        hell = State(str(hell_id))
        while hell in dfa.states:
            hell_id += 1
            hell.name = str(hell_id)
        dfa.states.add(hell)

        for state in dfa.states:
            for character in dfa.characters:
                if (state, character) in dfa.transition:
                    total[state, character] = dfa.transition[state, character]
                else:
                    total[state, character] = hell

        dfa.transition = total
        return dfa

    def is_total(self) -> bool:
        for state in self.states:
            for character in self.characters:
                if (state, character) not in self.transition:
                    return False
        return True

    def accepts(self, word: Union[str, List[Character]]) -> bool:
        if isinstance(word, str):
            evaluate = []
            for letter in word:
                character = Character(letter)
                evaluate.append(character)
        else:
            evaluate = word

        state = self.init
        for character in evaluate:
            if (state, character) not in self.transition:
                return False
            state = self.transition[state, character]
        return state in self.final

    def complement(self) -> DFA:
        dfa = deepcopy(self)
        if not dfa.is_total():
            dfa = dfa.total()

        dfa.final = dfa.states.difference(dfa.final)
        return dfa

    @staticmethod
    def composition(dfa_1: DFA, dfa_2: DFA, operation: Composition) -> DFA:

        if not dfa_1.is_total():
            dfa_1 = dfa_1.total()
        if not dfa_2.is_total():
            dfa_2 = dfa_2.total()

        # new characters
        characters = dfa_1.characters.union(dfa_2.characters)
        # ATTENTION: terminals may differ -> make total at the end?

        # new init
        init = State(f"{dfa_1.init.name}_{dfa_2.init.name}")

        # new states and transition
        states = {init}
        transition: Dict[Tuple[State, Character], State] = dict()
        for state_1 in dfa_1.states:
            for state_2 in dfa_2.states:
                state = State(f"{state_1.name}_{state_2.name}")  # TODO avoid conflicts!
                states.add(state)

                # new transition
                for character in characters:
                    if (state_1, character) in dfa_1.transition and \
                            (state_2, character) in dfa_2.transition:
                        dest_state_1 = dfa_1.transition[state_1, character]
                        dest_state_2 = dfa_2.transition[state_2, character]
                        dest_state = State(dest_state_1.name + "_" + dest_state_2.name)
                        transition[state, character] = dest_state

        # new final
        final: Set[State] = set()
        for state_1 in dfa_1.final:
            for state_2 in dfa_2.final:
                state = State(f"{state_1.name}_{state_2.name}")

                if operation == Composition.Union:
                    final.add(state)

                elif operation == Composition.Intersection and \
                        state_1 in dfa_1.final and \
                        state_2 in dfa_2.final:
                    final.add(state)

                elif operation == Composition.Subtraction and \
                        state_2 not in dfa_2.final:
                    final.add(state)

        dfa = DFA(states, characters, transition, init, final)
        return dfa

    @staticmethod
    def union(dfa_1: DFA, dfa_2: DFA) -> DFA:
        return DFA.composition(dfa_1, dfa_2, Composition.Union)

    @staticmethod
    def intersection(dfa_1: DFA, dfa_2: DFA) -> DFA:
        return DFA.composition(dfa_1, dfa_2, Composition.Intersection)

    @staticmethod
    def subtraction(dfa_1: DFA, dfa_2: DFA) -> DFA:
        return DFA.composition(dfa_1, dfa_2, Composition.Subtraction)

    def eliminate_unreachable(self) -> DFA:
        dfa = deepcopy(self)

        previous: Set[State] = set()
        actual = {self.init}

        while previous != actual:
            previous = deepcopy(actual)

            for state in previous:
                for character in dfa.characters:
                    if (state, character) in self.transition:
                        actual.add(self.transition[state, character])

        dfa.states = actual
        dfa.final = dfa.final.intersection(actual)
        for (state, character) in self.transition:
            if state not in actual:
                del dfa.transition[state, character]

        return dfa

    def sort_characters(self) -> List[Character]:
        return sorted(self.characters, key=lambda x: x.name)

    def minimize(self) -> DFA:
        dfa = deepcopy(self)
        dfa = dfa.eliminate_unreachable()
        dfa = dfa.total()

        characters = self.sort_characters()  # bc I want to iterate always in same order

        previous: Optional[Dict[State, int]] = None
        actual = dict()
        for state in dfa.states:
            if state in dfa.final:
                actual[state] = 1
            else:
                actual[state] = 0

        while previous != actual:
            previous = deepcopy(actual)
            patterns: Dict[Tuple[int, ...], Set[State]] = dict()
            actual = dict()

            for state in dfa.states:
                pattern = [previous[state]]
                for character in characters:
                    pattern.append(previous[dfa.transition[state, character]])

                if tuple(pattern) in patterns:
                    patterns[tuple(pattern)].add(state)
                else:
                    patterns[tuple(pattern)] = {state}

            i = 0
            for states in patterns.values():
                for state in states:
                    actual[state] = i
                i += 1

        new_transition: Dict[Tuple[State, Character], State] = {}
        dfa.states = set()
        dfa.final = set()
        for state in actual:
            new_state = State("new_" + str(actual[state]))
            dfa.states.add(new_state)

            for character in characters:
                if not (new_state, character) in new_transition:
                    old = dfa.transition[state, character]
                    now = State("new_" + str(actual[old]))
                    new_transition[new_state, character] = now

            if state == self.init:
                dfa.init = new_state
            if state in self.final:
                dfa.final.add(new_state)

        dfa.transition = new_transition
        return dfa

    def rename_state(self, rename: State, name: str):
        new_state = State(name)
        self.states.add(new_state)
        self.states.remove(rename)
        transition = deepcopy(self.transition)

        if rename == self.init:
            self.init = new_state
        if rename in self.final:
            self.final.add(new_state)
            self.final.remove(rename)

        for character in self.characters:
            if (rename, character) in self.transition:
                transition[new_state, character] = self.transition[rename, character]
                del transition[rename, character]

        for state, character in transition:
            if transition[state, character] == rename:
                del transition[state, character]
                transition[state, character] = new_state

        self.transition = transition

    def canonize(self) -> DFA:
        # TODO total?

        dfa = deepcopy(self)
        remain = deepcopy(self.states)
        characters = dfa.sort_characters()

        i = 0
        queue = [dfa.init]  # TODO deque
        while len(queue) > 0:
            actual = queue.pop(0)

            for character in characters:
                state = dfa.transition[actual, character]
                if state in remain:
                    queue.append(state)
                    remain.remove(state)

            dfa.rename_state(actual, str(i))
            i += 1
        return dfa

    def dfa_to_reg(self) -> Any:
        from reg_grammars import REG, Terminal, Nonterminal

        nonterminals: Set[Nonterminal] = set()
        terminals: Set[Terminal] = set()
        rules: Dict[Nonterminal, Set[Union[Terminal, Tuple[Terminal, Nonterminal]]]] = dict()

        for state in self.states:
            nonterminal = Nonterminal(state.name)
            nonterminals.add(nonterminal)

        for character in self.characters:
            terminal = Terminal(character.name)
            terminals.add(terminal)
        init = Nonterminal("new_init")  # TODO assert that name is unique, as for hell in make_total
        init = init
        nonterminals.add(init)

        for state, character in self.transition:
            dest_state = self.transition[state, character]
            if state == self.init:
                nonterminal2 = Nonterminal(dest_state.name)
                terminal = Terminal(character.name)
                if init in rules:
                    rules[init].add((terminal, nonterminal2))
                else:
                    rules[init] = {(terminal, nonterminal2)}
                if dest_state in self.final:
                    rules[init].add(terminal)
            else:
                nonterminal1 = Nonterminal(state.name)
                nonterminal2 = Nonterminal(dest_state.name)
                terminal = Terminal(character.name)
                if nonterminal1 in rules:
                    rules[nonterminal1].add((terminal, nonterminal2))
                else:
                    rules[nonterminal1] = {(terminal, nonterminal2)}
                if dest_state in self.final:
                    rules[nonterminal1].add((terminal))

        reg = REG(nonterminals, terminals, rules, init)
        if self.init in self.final:
            pass  # TODO eg.rules[init].add((\eps))

        return reg

    def is_empty(self) -> bool:
        reached: Deque[State] = deque([self.init])
        reachable: Set[State] = {self.init}
        predecessor: Dict[State, Tuple[State, Character]] = {}

        while len(reached) > 0:
            actual = reached.popleft()
            for character in self.characters:
                if (actual, character) in self.transition:
                    dest_state = self.transition[actual, character]
                    if dest_state not in reachable:
                        reached.append(self.transition[actual, character])
                        reachable.add(self.transition[actual, character])
                        predecessor[dest_state] = (actual, character)

        if len(reachable.intersection(self.final)) == 0:
            return IsEmptyResult()
        else:
            word = []
            state = reachable.intersection(self.final).pop()
            while state is not self.init:
                word.append(predecessor[state][1].name)
                state = predecessor[state][0]

            counterexample = IsEmptyResult(''.join(reversed(word)))
            return counterexample

    # Ha!
    def is_universal(self):
        return self.complement().is_empty()

    # returns: True or (1. what IS in first language (self) and NOT in second (dfa), 2. analogically)
    def is_equivalent(self, dfa: DFA) -> IsEquivalentResult:
        left_empty = self.intersection(self, dfa.complement()).is_empty()
        right_empty = self.intersection(self.complement(), dfa).is_empty()

        if left_empty and right_empty:
            result = IsEquivalentResult()
        else:
            result = IsEquivalentResult(left_empty.counterexample, right_empty.counterexample)
        return result

    def is_part_identical(self, dfa: DFA) -> bool:
        return self.states == dfa.states and self.characters == dfa.characters and \
               self.transition == dfa.transition and self.init == dfa.init and \
               self.final == dfa.final

    def is_minimal(self) -> bool:
        minimal = self.minimize()
        return len(self.states) == len(minimal.states) and \
               len(self.transition) == len(minimal.transition)

    def is_canonical(self) -> bool:
        canonic = self.canonize()
        return self.is_part_identical(canonic)

class NFA:
    Transition = Dict[Tuple[State, Character], Set[State]]
    type_var = TypeVar('type_var')

    def __init__(self, states: Set[State],
                 characters: Set[Character],
                 transition: Transition,
                 init: State,
                 final: Set[State]):

        self.states = states
        self.characters = characters
        self.transition = transition
        self.init = init
        self.final = final

    def determinize(self) -> DFA:
        states: Set[FrozenSet[State]] = set()
        states.add(frozenset({self.init}))
        transition = {}
        final = set()
        done: Set[FrozenSet[State]] = set()

        while len(states.difference(done)) > 0:
            subset = (states.difference(done)).pop()  # arbitrary element from set
            if len(subset.intersection(self.final)) > 0:
                final.add(subset)
            for character in self.characters:
                new_subset: Set[State] = set()
                for state in subset:
                    if (state, character) in self.transition:
                        new_subset = new_subset.union(self.transition[state, character])

                states.add(frozenset(new_subset))
                transition[subset, character] = frozenset(new_subset)

            done.add(subset)

        new_states: Set[State] = set()
        new_transition: Dict[Tuple[State, Character], State] = dict()
        new_final: Set[State] = set()

        for state_set in states:
            new_states.add(self.unset(state_set))
        for state_set in final:
            new_final.add(self.unset(state_set))
        for state_set, character in transition:
            new_transition[self.unset(state_set), character] = self.unset(transition[state_set, character])

        dfa = DFA(new_states, self.characters, new_transition, self.init, new_final)
        return dfa

    def unset(self, states: FrozenSet[State]) -> State:
        return State('_'.join(set(map(lambda x: x.name, sorted(states, key=lambda x: x.name)))))

class NFAE:
    Transition = Dict[Tuple[State, Union[Character, Eps]], Set[State]]
    type_var = TypeVar('type_var')

    def __init__(self, states: Set[State],
                   characters: Set[Character],
                   transition: Transition,
                         init: State,
                        final: Set[State]):
            self.states = states
            self.characters = characters
            self.transition = transition
            self.init = init
            self.final = final

    def eliminate_epsilon(self) -> NFA:

        surroundings: Dict[State, Set[State]] = {}
        for state in self.states:
            surroundings[state] = self.epsilon_surroundings(state)

        new_transition: Dict[Tuple[State, Character], Set[State]] = {}
        for state in self.states:
            for character in self.characters:
                reached_states = set()
                if (state, character) in self.transition:
                    reached_states.update(self.transition[state, character])

                for eps_state in surroundings[state]:
                    if (eps_state, character) in self.transition:
                        reached_states.update(self.transition[eps_state, character])

                if len(reached_states) > 0:
                    new_transition[state, character] = deepcopy(reached_states)
                    for reached in reached_states:
                        new_transition[state, character].update(surroundings[reached])

        new_final = deepcopy(self.final)
        if surroundings[self.init].intersection(self.final):
            new_final.add(self.init)

        nfa = NFA(self.states, self.characters, new_transition, self.init, new_final)
        return nfa

    def epsilon_surroundings(self, state: State) -> Set[State]:
        reached: Deque[State] = deque([state])
        reachable: Set[State] = {state}
        eps = Eps()

        while len(reached) > 0:
            actual = reached.popleft()
            if (actual, eps) in self.transition:
                for dest_state in self.transition[actual, eps]:
                    if dest_state not in reachable:
                        reached.append(dest_state)
                        reachable.add(dest_state)

        return reachable